from base64 import urlsafe_b64decode, urlsafe_b64encode
import hashlib

from . import utils


class DecoderRing:
    """
    Many options for encoding / decoding.
    Also includes some encryption options such as MD5.
    """
    def __init__(self, data: str) -> None:
        self.data: str = data
        self.altered_data: str = ''

        if not isinstance(data, str):
            raise TypeError("Input 'data' requires type 'str'.")

    def __call__(self):
        return {
            "data": self.data,
            "altered_data": self.altered_data
        }

    def _data(self) -> str:
        """
        Use self.altered_data over self.data
        """
        if self.altered_data:
            return self.altered_data
        return self.data

    def base64_encode(self) -> str:
        """
        Encode into Base64 string
        """
        data = self._data()
        encoded = urlsafe_b64encode(data.encode('utf-8'))
        self.altered_data = str(encoded, 'utf-8')
        return self.altered_data

    def base64_decode(self) -> str:
        """
        Decode Base64 string into ASCII
        """
        data = self._data()
        decoded = urlsafe_b64decode(data)
        self.altered_data = decoded.decode()
        return self.altered_data

    def hex_encode(self) -> str:
        """
        Encode into Hexadecimal string
        """
        data = self._data()
        bytes_conversion = data.encode('utf-8')
        self.altered_data = bytes_conversion.hex()
        return self.altered_data

    def hex_decode(self) -> str:
        """
        Decode Hexadecimal into ASCII
        """
        data = self._data()
        self.altered_data = bytes.fromhex(data).decode('utf-8')
        return self.altered_data

    def caeser_encode(self, shift: int = 5) -> str:
        """
        Caeser's Cipher used on a string.
        """
        data = self._data()

        result = ""

        for char in data:
            if utils.is_not_ascii_letter(character=char):  # Account for spaces and special chars.
                result += char
            elif char.isupper():
                enciphered_index = (utils.ascii_to_index(letter=char) + shift) % 26
                result += utils.index_to_ascii(index=enciphered_index, capitalize=True)
            else:
                enciphered_index = (utils.ascii_to_index(letter=char) + shift) % 26
                result += utils.index_to_ascii(index=enciphered_index)
        return result

    def caeser_decode(self, shift: int = 5) -> str:
        """
        Decipher Caeser's Cipher. Shift should be the value that was used to originally encipher the text.
        """
        data = self._data()

        result = ""

        for char in data:
            if utils.is_not_ascii_letter(character=char):  # Account for spaces and special chars.
                result += char
            elif char.isupper():
                enciphered_index = (utils.ascii_to_index(letter=char) - shift) % 26
                result += utils.index_to_ascii(index=enciphered_index, capitalize=True)
            else:
                enciphered_index = (utils.ascii_to_index(letter=char) - shift) % 26
                result += utils.index_to_ascii(index=enciphered_index)
        return result

    def encrypt_md5(self):
        data = self._data()
        hashed_value = hashlib.md5(data.encode())
        return hashed_value.hexdigest()



import socket
from select import select
from threading import Thread

from .listener import Server


class SummoningCircle(Server):
    """
    NOTE: Currently, only works with SummoingCircleInterpreter. Separate package.
    Interpreter used for SummoningCircleServer SummoningCircle.
    """
    def __init__(self):

        super().__init__()

        self.command_prefix: str = '--'
        self.buffer_size: int = 1024 * 25
        self.carriage_return: str = '\n'
        self.timeout_in_sec: int = 5

        self._basic_functions = {
            f"{self.command_prefix}shutdown": self.shutdown,
            f"{self.command_prefix}exit": self.exit,
            f"{self.command_prefix}view": self.view_clients
        }

        self._advanced_functions = {
            f"{self.command_prefix}name": self.change_nickname,
            f"{self.command_prefix}session": self.change_session
        }

    def _parser(self, *args, **kwargs) -> list:
        """
        Parse used to set up decision-making.
        """

        args_list = list(args)
        command = args_list.pop(0)

        if len(args) > 1:
            return [command, args_list]
        return [command, None]

    def _set_active_session(self):
        """
        Used to ensure 'self.active_session' is using the correct client socket.
        """
        if self.clients:
            self.active_session = self.clients[self.active_client_index].get('client_socket')

    def change_nickname(self, interpreter_options: list[int, str]):
        """
        Command: Change 'client_nickname' for a specified index in 'self.clients'
        """
        if len(interpreter_options) > 1:
            index = int(interpreter_options[0])
            name_change = interpreter_options[1]
            self.clients[index]['client_nickname'] = name_change
            return f'Name changed: {name_change}'

    def change_session(self, interpreter_options: list[int]):
        """
        Command: Change client session
        """

        self.active_client_index = int(interpreter_options[0])
        return f'Session changed: {self.active_client_index}'

    def exit(self):
        """
        Command: Exit client connection and close connection.
        Also, clean up 'self.clients' list.
        """
        try:
            # Create new list excluding current client_socket
            self.clients = [
                client for index, client in enumerate(self.clients) if self.active_session != client['client_socket']
            ]

            self.active_session.shutdown(socket.SHUT_RDWR)
            self.active_session.close()
        except (socket.error, OSError, ValueError):
            pass

    def is_active(self):
        """
        Command: Basic ping check to test if client is active.
        Works for basic connections but not a viable solution for shell / SummoningCircle concept.
        """
        try:
            self.active_session.send(b'echo ping\n')
            return True
        except socket.error:
            return False

    def send_command(self, command: str) -> str:
        """
        Command: Default way to send OS commands to the active connection's shell.
        """
        self.active_session.send(f"{command}{self.carriage_return}".encode())

        # Confirm data is being sent with a timeout
        ready = select([self.active_session], [], [], self.timeout_in_sec)

        if ready[0]:
            return self.active_session.recv(self.buffer_size).decode()

        return 'Response timed Out.'

    def shutdown(self):
        """
        Command: Close client connections and shutdown server.
        Finally, system exit.
        """
        try:
            for client in self.clients:
                client['client_socket'].close()
            self.server.shutdown(socket.SHUT_RDWR)
        except (socket.error, OSError, ValueError):
            pass
        self.server.close()
        exit()

    def start_listener(self):
        """
        Startup SummoningCircle in the background.
        Start interpreter.
        """
        listener = Thread(target=self.bind_socket)
        listener.start()
        return "Server started"

    def view_clients(self):
        """
        Command: View all client connections in 'self.clients'
        """
        output = '[*] Summoned Connections\n\n'
        for index, client in enumerate(self.clients):
            output += f"{index} -> {client['client_nickname']} : {client['client_address']}\n"
        return output

    def instruction(self, *args, **kwargs):
        """
        Used for decision making. Called by main interpreter function.
        """

        if 'start' == args[0].lower():
            return self.start_listener()

        if not self.clients:
            # empty command or no clients
            return "No clients"

        self._set_active_session()

        interpreter_command, interpreter_options = self._parser(*args, **kwargs)

        if interpreter_command in self._basic_functions.keys():
            function = self._basic_functions.get(interpreter_command)
            print(function())
        elif interpreter_command in self._advanced_functions.keys():
            function = self._advanced_functions.get(interpreter_command)
            print(function(interpreter_options))
        else:
            command = ' '. join(list(args))
            print(self.send_command(command=command))



import socket
from threading import Thread
from time import sleep


class Server:
    """
    Listener designed to work with Netcat reverse shells.
    """
    def __init__(
            self,
            local_host: str = '0.0.0.0',
            local_port: int = 5000,
    ):

        self.local_host = local_host
        self.local_port = local_port

        self.server: socket = None
        self.active_session: socket = None

        self.active_client_index: int = 0
        self.active_session: socket = None

        self.clients = []
        """
        self.clients = [{
            "client_nickname": str,
            "client_address": str,
            "client_socket": socket
        }]
        """

    def bind_socket(self) -> None:
        """
        Start listening socket server.
        """
        try:

            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.bind((self.local_host, self.local_port))
            self.server.listen(5)

            print(f"Listening for connection(s)...")

            while True:
                client_socket, client_address = self.server.accept()
                self.active_session = client_socket

                # Inherited self.clients from Interpreter
                self.clients.append({
                    "client_nickname": client_address,
                    "client_address": client_address,
                    "client_socket": client_socket
                })

                print(f"Connection: {client_address[0]}:{client_address[1]}")

        except (socket.error, KeyboardInterrupt):
            print(f'[-] Problem encountered: {socket.error} ')
